;; Generated package description from phps-mode.el  -*- no-byte-compile: t -*-
(define-package "phps-mode" "0.3.65.0.2020102191115" "Major mode for PHP with Semantic integration" '((emacs "26")) :keywords '("tools" "convenience") :authors '(("Christian Johansson" . "christian@cvj.se")) :maintainer '("Christian Johansson" . "christian@cvj.se") :url "https://github.com/cjohansson/emacs-phps-mode")
